import math
import multiprocessing
from typing import Dict, Generator, List, Optional, Tuple

import pandas
from loguru import logger
from tqdm import tqdm

try:
	from muller.clustering.metrics import distance_methods
	from muller import widgets
except ModuleNotFoundError:
	from . import distance_methods
	from ... import widgets

FilterType = Tuple[Optional[pandas.Series], Optional[pandas.Series]]


class DistanceCalculator:
	""" Refactor to clean up code and add multithreading."""

	def __init__(self, detection_limit: float, fixed_limit: float, metric: str, threads: Optional[int] = None):
		self.detection_limit = detection_limit
		self.fixed_limit = fixed_limit
		self.metric = metric
		self.threads = threads
		# Basically used as a cache. Should save memory compared to loading each pair of trajectories directly into `pair_combinations`.
		self.trajectories: Optional[pandas.DataFrame] = None

		self.progress_bar_minimum_points = 10000  # The value to activate the scale bar at.

	def calculate_pairwise_distances_threaded(self, pair_combinations: Generator, total: Optional[int] = None) -> Dict[
		Tuple[str, str], float]:
		""" Threaded version of the pairwise distance calculator"""
		pair_array: Dict[Tuple[str, str], float] = dict()
		progress_bar = tqdm(total = total)
		pool = multiprocessing.Pool(processes = self.threads)
		for i in tqdm(
				[pool.apply_async(calculate_distance, args = (self, e, self.trajectories)) for e in pair_combinations]):
			key, value = i.get()  # retrieve the calculated value
			pair_array[key] = value
			pair_array[key[::-1]] = value
			progress_bar.update(1)

		return pair_array

	def calculate_pairwise_distances_serial(self, pair_combinations: Generator, total: Optional[int] = None):
		""" Nonthreaded version of the pairwise distance calculator"""
		pair_array: Dict[Tuple[str, str], float] = dict()

		# The progressbar is not really useful if there aren;t a lot of combinations, since calculating the pairwise
		# distances is pretty fast. So disable the progressbar when the expected time to calculate all distances
		# Is less than, say, 20-30s.
		use_progressbar = total >= 10_000
		if use_progressbar:
			progress_bar = tqdm(total = total)

		for element in pair_combinations:
			key, value = calculate_distance(self, element, self.trajectories)
			pair_array[key] = value
			pair_array[key[1], key[
				0]] = value  # It's faster to add the reverse key rather than trying trying to get  test forward and reverse keys
			if use_progressbar:
				progress_bar.update(1)
		return pair_array

	def calculate_pairwise_distances(self, labels: List[str]) -> Dict[Tuple[str, str], float]:
		""" Implements the actual loop over all pairs of trajectories.
		"""
		# May as well move the combination function here so we don't have to pass an additional parameter specifying the total number
		# of trajectories so tqdm workd properly.

		total_elements = len(labels)
		total_combinations = widgets.calculate_number_of_combinations(total_elements)

		logger.debug(
			f"Generating pairwise combinations with {total_elements} items resulting in {total_combinations} combinations...")

		if total_combinations > 100_000_000:
			message = f"The provided dataset has {total_elements} trajectories, which requires {total_combinations} distance calculations." \
					  "This will require a long time to process (you may need to adjust the number of available threads with the --threads option)" \
					  "and will consume a large amount of memory (i.e. more than 16GB)."
			logger.warning(message)
		pair_combinations = widgets.get_pair_combinations(labels)

		if self.threads and self.threads > 1:  # One process is slower than using the serial method.
			logger.debug(f"Using multithreading...")
			pair_array = self.calculate_pairwise_distances_threaded(pair_combinations, total_combinations)
		else:
			logger.debug(f"Using a single thread...")
			pair_array = self.calculate_pairwise_distances_serial(pair_combinations, total_combinations)

		# Assume that any pair with NAN values are the maximum possible distance from each other.
		try:
			maximum_distance = max(filter(lambda s: not math.isnan(s), pair_array.values()))
		except ValueError:
			# Usually cause by max() being used on an empty sequence.
			message = f"Could not calculate the pairwise distances due to invalid series (usually because all measurements are below the detectionlimit"
			raise ValueError(message)

		pair_array = {k: (v if not math.isnan(v) else maximum_distance) for k, v in pair_array.items()}

		return pair_array

	def run(self, trajectories: pandas.DataFrame) -> Dict[Tuple[str, str], float]:
		"""
			Calculates the distance between all pairwise combinations of mutational trajectories. The total number of pairs can be calculated by
			`n!/(n-2)!`, which can be simlified to `n*(n-1)`.
			10 trajectories -> 90
			100 trajectories -> 9900
			1000 trajectories -> 999000
			31000 trajectories -> 9.61E8

		Parameters
		----------
		trajectories
		"""
		logger.debug("Calculating the pairwise values...")
		logger.debug(f"\t detection limit: {self.detection_limit}")
		logger.debug(f"\t fixed limit: {self.fixed_limit}")
		logger.debug(f"\t metric: {self.metric}")
		logger.debug(f"\t threads: {self.threads}")

		self.trajectories = trajectories

		pairwise_distances = self.calculate_pairwise_distances(trajectories.index)

		return pairwise_distances


def get_pair_category(left: pandas.Series, right: pandas.Series, dlimit: float, flimit: float) -> str:
	"""
		Categorizes the pair of mutational trajectories based on the dynamics of the
		measured frequencies over time.
		Each pair of mutational trajectories can be categorized into one of the following groups:
		- "onlyFixed": Both trajectories were only detected during timepoints where they were fixed.
		- "PartiallyFixed":Both trajectories were fixed, but only one trajectory was fixed during all timepoints where it was detected.
		- "bothFixed": Both trajectories where fixed during at least one sampled timepoint, and both had intermediate values.
		- "oneFixed": Only one of the trajectories was fixed.
		- "notFixed: Neither trajectory where fixed at any of the sampled timepoints.
	Returns
	-------
	str
		- 'onlyFixed': Both trajectories were only detected after they fixed.
		- 'partiallyFixed': Only one trajectory had intermediate values.
		- 'bothFixed': Both trajectories had non-fixed timepoints.
		- 'oneFixed': Only one trajectory ever fixed.
		- 'notFixed': Neither trajectory was ever fixed.
	"""
	fixed_left = widgets.get_fixed(left, flimit)
	fixed_right = widgets.get_fixed(right, flimit)

	left_was_fixed = len(fixed_left) > 0
	right_was_fixed = len(fixed_right)
	# Check if there are any intermediate values between the detection limit and the fixed limit.
	intermediate_left = widgets.get_intermediate(left, dlimit, flimit)
	intermediate_right = widgets.get_intermediate(right, dlimit, flimit)

	#undetected_left = widgets.get_undetected(left, dlimit)
	#undetected_right = widgets.get_undetected(right, dlimit)

	if left_was_fixed and right_was_fixed:
		# If both of these fixed at some point, the regions where they're both fixed means they should be
		# included within the same genotype. If, however, they both are both fixed but at separate timepoints,
		# They probably shouldn't be grouped together. It is also possible for one to be fixed and the other to
		# be only fixed.
		left_was_only_fixed = intermediate_left.empty
		right_was_only_fixed = intermediate_right.empty

		if left_was_only_fixed and right_was_only_fixed:
			# Both series only have undetected or fixed values
			category = 'onlyFixed'
		elif not left_was_only_fixed and not right_was_only_fixed:
			# Both series had intermediate values
			category = 'bothFixed'
		else:
			# One series only had fixed timepoints, the other did not.
			category = 'partiallyFixed'

		# So, check if the 'fixed' regions overlap.
		# left_reduced, right_reduced = widgets.get_valid_points(left_trajectory, right_trajectory, process.detection_limit, process.fixed_limit,
		#	inner = False)
		overlapping_regions = widgets.get_overlap_regions(left, right, flimit)
		# Need to make sure there aren't any regions after the first timepoint where both are fixed that is not fixed
		first_overlap_timepoint = widgets.get_first_fixed_timepoint(overlapping_regions, flimit)
		# Now get the section of the series after the first fixed timepoint for both series.
		overlapping_regions_after_fixed: pandas.Series = overlapping_regions[first_overlap_timepoint:]
		# Are there any timepoints after this where they aren't both overlapping?
		nonfixed_regions_after_first = overlapping_regions.apply(lambda s: s < 0.97)
		# If there is no subsequent timepoint that is not fixed or the overlapping regions after fixed is empty
		# Then these should be grouped together, but still need to check if the nonfixed regions agree.
		#is_genotype = overlapping_regions_after_fixed.empty or nonfixed_regions_after_first.sum() == 0

	# Check if the number of timepoints that overlap is more than one (overlapping suring one timepoint may be
	# due to measurement error.
	elif (left_was_fixed and not right_was_fixed) or (not left_was_fixed and right_was_fixed):
		# Only one trajectory had fixed timepoints
		category = 'oneFixed'
	elif not left_was_fixed and not right_was_fixed:
		# Neither series had fixed values.
		category = 'notFixed'
	else:
		message = f"Could not determine the category of this mutational pair"
		logger.error(message)
		logger.error(left)
		logger.error(right)
		raise ValueError

	return category


def filter_timepoints(left_trajectory: pandas.Series, right_trajectory: pandas.Series, dlimit: float,
		flimit: float) -> FilterType:
	"""
		Filters the available timepoints based on the measured dynamics.
	"""
	"""
	Legacy code:
	if left_was_fixed == right_was_fixed:
		left_reduced, right_reduced = widgets.get_valid_points(left_trajectory, right_trajectory, process.detection_limit, process.fixed_limit,
			inner = False)
	else:
		left_reduced, right_reduced = widgets.get_valid_points(left_trajectory, right_trajectory, process.detection_limit, inner = False)

	
	"""
	pair_category = get_pair_category(
		left_trajectory, right_trajectory,
		dlimit = dlimit, flimit = flimit
	)
	if pair_category == 'onlyFixed':
		left_reduced = right_reduced = None

	elif pair_category == 'partiallyFixed':
		# There is no overlap between these series so we have to rely on the overlap between "fixed" regions.
		left_reduced, right_reduced = widgets.get_valid_points(
			left_trajectory, right_trajectory,
			dlimit = dlimit, flimit = flimit, inner = False
		)

	elif pair_category == 'oneFixed':
		left_reduced, right_reduced = widgets.get_valid_points(
			left_trajectory, right_trajectory,
			dlimit = dlimit, inner = False
		)
	elif pair_category == 'notFixed':
		left_reduced, right_reduced = widgets.get_valid_points(
			left_trajectory, right_trajectory,
			dlimit = dlimit, flimit = flimit,
			inner = False
		)
	elif pair_category == 'bothFixed':
		left_reduced, right_reduced = widgets.get_valid_points(
			left_trajectory, right_trajectory,
			dlimit = dlimit, flimit=flimit,
			inner = False
		)
	else:
		message = f"Got an invalid category for a pair of trajectories: '{pair_category}'"
		logger.error(message)
		logger.error(left_trajectory.tolist())
		logger.error(right_trajectory.tolist())
		raise ValueError(message)
	return left_reduced, right_reduced


# Keep this as a separate function. Class methods are finicky when used with multiprocessing.
def calculate_distance(process: DistanceCalculator, element: Tuple[str, str], trajectories: pandas.DataFrame) -> Tuple[
	Tuple[str, str], float]:
	""" Implements the actual calculation for a specific pair of trajectories.
		It should be atomitized so that it works with multithreading.
	"""
	left, right = element
	left_trajectory = trajectories.loc[left]
	right_trajectory = trajectories.loc[right]

	# We only care about the timepoints such that `detection_cutoff` < f < `fixed_cutoff`.
	# For now, lets require that both timepoints are detected and not yet fixed.
	# There is an issue related to comparing fixed genotypes against non-fixed genotypes.

	left_reduced, right_reduced = filter_timepoints(
		left_trajectory, right_trajectory, process.detection_limit, process.fixed_limit
	)

	if left_reduced is None or right_reduced is None:
		# Treat both trajectories as fixed immediately.
		distance_between_series = fixed_overlap(left_trajectory, right_trajectory, process.fixed_limit)
	else:
		distance_between_series = distance_methods.calculate_distance(left_reduced, right_reduced, process.metric)

	return element, distance_between_series


def fixed_overlap(left: pandas.Series, right: pandas.Series, fixed_cutoff: float) -> float:
	"""
		Calculates the overlap of two series that are both undetected or fixed at all timepoints.
		This will be the distance value for these two series.
	Parameters
	----------
	left:pandas.Series
	right:pandas.Series
	fixed_cutoff: float
	"""
	left_fixed: pandas.Series = left[left.ge(fixed_cutoff)].dropna()
	right_fixed: pandas.Series = right[right.ge(fixed_cutoff)].dropna()
	if left_fixed.empty or right_fixed.empty:
		# Both are undetected, since they both failed the invalid timepoint filter.
		# As in, was already tested for intermediate values, so the only timepoints left would be undetected or fixed,
		# and these were just tested for fixed values. However, this situation is usually filtered during the earlier
		# steps so don't expect this if branch to be run.
		result = False
	else:
		# Since these genotypes fixed immediately, they should only be grouped together if they
		# fixed at the same timepoint.
		fixed_at_same_time = left_fixed.index[0] == right_fixed.index[0]
		# Check if the trajectories overlap completely
		overlap = set(left_fixed.index) & set(right_fixed.index)
		# This could probably be replaced with set(a) == set(b), but for some reason I compare them to
		# The union instead. It will do the same job and seems to work so I'll leave it for now.
		complete_overlap = len(overlap) == len(left_fixed) and len(overlap) == len(right_fixed)
		result = fixed_at_same_time and complete_overlap
	result = 0 if result else math.nan

	return result




import math
import re
from pathlib import Path
from typing import Any, Optional, Tuple, Union

import pandas
from loguru import logger

try:
	from muller.widgets import get_numeric_columns
	from muller.dataio import import_table
except ModuleNotFoundError:
	from ..widgets import get_numeric_columns
	from ..dataio import import_table


def _convert_to_integer(value: Any, default: Optional[int] = None) -> int:
	""" Attempts to convert the input value to an integer. Returns `default` otherwise."""
	if isinstance(value, str) and (value.startswith('x') or value.startswith('X')):
		value = value[1:]
	try:
		result = int(value)
	except (TypeError, ValueError):
		result = default
	return result


def _correct_math_scale(old_data: pandas.DataFrame) -> pandas.DataFrame:
	""" Ensures the time table columns contain values between 0 and 1 and are of type `float`"""
	new_data = old_data.copy(deep = True)
	for column in old_data.columns:
		old_column = old_data[column]
		try:
			maximum_value = old_column.max()
			minimum_value = old_column.min()
		except TypeError as exception:
			logger.error(f"Some of the values in the column '{column}' could not be read as numbers.")
			logger.error(f"The column had values of {old_data[column].tolist()}")
			raise exception

		# Test the most common case first
		if (0.0 <= minimum_value <= 1.0) and (0.0 <= maximum_value <= 1.0):
			# The table is fine.
			new_column = old_column.astype(float) # To match the other columns.
		elif maximum_value > 2.0:
			# A little fuzzy here in case the max is something like 1.01. This should be masked instead.
			# This column uses a 0-100 range. convert to 0-1
			logger.warning(f"The column `{column}` had values greater than 1.0. It will be converted to a float between 0 and 1.")
			new_column = old_column / 100
		else:
			# The table is formatted strangely, likely due to human error.
			# Try to coerce the column values into the 0-1 range.
			# Basically, mask values that exceed the lower and upper limits.
			below_o = (old_column < 0).sum()
			above_1 = (old_column > 1).sum()
			message = f"The column '{column}' had values outside the range [0,1], but not in the range [0,100]. It will be coerced to [0,1]."
			message += f"There were {below_o} negative values and {above_1} values above 1.0."
			logger.warning(message)

			new_column = old_column.mask(lambda s: s < 0, 0)
			new_column = new_column.mask(lambda s: s > 1, 1)

		new_data[column] = new_column
	return new_data


def convert_string_to_number(value: str) -> float:
	pattern = "[\d]+(?:[.][\d]+)?"
	if isinstance(value, str):
		match = re.search(pattern, value)
		try:
			result = float(match.group(0))
		except (AttributeError, ValueError):
			# AttributeError: No matches were found
			# ValueError: THe matched string could not be converted to a number
			logger.error(f"Could not convert '{value}' to a number.")
			result = math.nan
	else:
		result = value
	return result


def _fix_column_datatypes(table: pandas.DataFrame) -> pandas.DataFrame:
	# Try to extract numerical data from the table.

	for column in table.columns:
		converted_column = table[column].apply(convert_string_to_number)
		table[column] = converted_column
	return table

def _add_key_column(table:pandas.DataFrame, key_column:str)->pandas.DataFrame:
	upper_limit = len(table)+1
	if key_column.lower() == "genotype":
		table[key_column] = [f"genotype-{i}" for i in range(1,upper_limit)]
	else:
		table[key_column] = list(range(1,upper_limit))

	return table

def _parse_table(raw_table: pandas.DataFrame, key_column: str) -> Tuple[pandas.DataFrame, pandas.DataFrame]:
	"""
		Converts column headers to integers and moves all non-integer columns to a separate dataframe.
	"""
	# Make sure the column with the series names is the index of the table.
	if key_column not in raw_table.columns:
		raw_table = _add_key_column(raw_table, key_column)
	raw_table = raw_table.sort_values(by = key_column)
	raw_table[key_column] = [str(i) for i in raw_table[key_column].tolist()]
	raw_table.set_index(key_column, inplace = True)
	# Extract the columns which indicate timepoints of observations. Should be integers.
	frequency_columns = get_numeric_columns(raw_table.columns)

	# Extract the columns with the trajectory identifiers and frequencies at each timepoint.
	time_table = raw_table[frequency_columns]
	# Convert the columns into integers. Makes them easier to work with if they have a standard raw_table type.
	converted_frequency_columns = [_convert_to_integer(i) for i in frequency_columns]
	time_table.columns = converted_frequency_columns
	# Sort the table columns
	time_table = time_table[sorted(time_table.columns)]
	try:
		# Test to see if the table was correctly converted to numerical datatypes.
		time_table[time_table.columns[0]].max()
	except TypeError:
		logger.warning(f"Some of the values could not be interpreted as numerical data. Attempting to extract numerical data from these values...")
		time_table = _fix_column_datatypes(time_table)

	# Drop any trajectories which are never detected.
	# noinspection PyUnresolvedReferences
	time_table = time_table[(time_table.T != 0).any()]

	# Make sure the table values are between 0 and 1 and are of type `float`
	time_table = _correct_math_scale(time_table)

	# Make sure the time table contains a column for timepoint `0`.
	if 0 not in time_table.columns:
		time_table[0] = 0.0  # Should be a float to match the dtype of the other columns
		logger.warning("Warning: The input table did not have values for timepoint 0. Adding 0% for each trajectory at timepoint 0")

	# Make sure there are no missing values.
	time_table = time_table.fillna(0)
	# Extract metadata for each series.
	info_table = raw_table[[i for i in raw_table.columns if i not in frequency_columns or i == key_column]]

	# Make sure that the columns in the time table are numeric and sorted properly.
	# Note that adding the '0' column above places it on the end on the table.
	time_table = time_table[sorted(time_table.columns)]

	return time_table, info_table


def parse_genotype_table(filename: Path, sheet_name: str = 'Sheet1') -> Tuple[pandas.DataFrame, pandas.DataFrame]:
	""" Imports a table that lists pre-computed genotypes rather than trajectories."""
	data = import_table(filename, sheet_name = sheet_name)
	# For some reason some tables are annotated with 'genotype   ' with extra spaces.
	data.columns = [(i.strip() if isinstance(i, str) else i) for i in data.columns]
	if 'Genotype' in data.columns:
		key_column = 'Genotype'
	elif 'Unnamed: 0' in data.columns:
		key_column = 'Unnamed: 0'
	elif 'Trajectory' in data.columns:
		key_column = 'Trajectory'
	else:
		message = f"One of the columns needs to be labeled `Genotype`. Got {data.columns} instead from {filename}."
		raise ValueError(message)

	genotype_timeseries, genotype_info = _parse_table(data, key_column)
	# Make sure the genotype labels are prefixed with 'genotype-'
	if not genotype_timeseries.index[0].startswith('genotype'):
		genotype_timeseries.index.name = 'originalLabel'
		genotype_timeseries['Genotype'] = [f'genotype-{i}' for i in range(1, len(genotype_timeseries) + 1)]
		genotype_timeseries = genotype_timeseries.reset_index()
		genotype_timeseries = genotype_timeseries.set_index('Genotype')
		genotype_timeseries.pop('originalLabel')

	# Try to sort the genotypes by label, if posible.
	# Note: designed to sort labels of the form `genotype-[\d]+`
	try:
		sorted_index = sorted(genotype_timeseries.index, key = lambda s: float(s.split('-')[-1]))
	except ValueError:
		sorted_index = genotype_timeseries.index

	genotype_timeseries = genotype_timeseries.loc[sorted_index]

	# Remove extraneous whitespace.
	return genotype_timeseries, genotype_info


def parse_trajectory_table(filename: Union[str, Path], sheet_name = 'Sheet1') -> Tuple[pandas.DataFrame, pandas.DataFrame]:
	"""
		Reads an excel or csv file. Assumes that the file has a `Trajectory` column and a column for each timepoint.
	Parameters
	----------
	filename: Path
		The table containing the trajectories and associated metadata. Can be an excel sheet or comma/tab delimited file.
	sheet_name: str; Default 'Sheet1'
		Indicates which sheet contains the data, if an excel table is given.
	Returns
	-------
	pandas.DataFrame, pandas.DataFrame
		A timeseries dataframe
			- Index -> str
				Names unique to each trajectory.
			- Columns -> int
				The timeseries points will correspond to the frequencies for each trajectory included with the input sheet.
				Each trajectory/timepoint will include the observed frequency at each timepoint.
		A metadata dataframe
			- Index -> str
				Identical index to the timeseries dataframe.
			- Columns -> str
				All columns from the original input table that do no correspond to timepoints.
	"""

	# Read in the data table.
	raw_data = import_table(filename, sheet_name)

	key_column = 'Trajectory'
	timeseries, info = _parse_table(raw_data, key_column)

	if 'genotype' in info:
		# This file was generated by a previous run.
		info.pop('genotype')
	# Make sure the index is named `Trajectory` for consistency
	timeseries.index.name = 'Trajectory'
	# Make sure the columns of `info` are lowercase to help with later parsing.
	info.columns = [i.lower() for i in info.columns]
	return timeseries, info



from typing import Dict, List, Union

import pandas


class GGMuller:
	""" Consolidates the functions used to save the ggmuller tables.
		Parameters
		----------
		cutoff_detection: float
		adjust_populations:bool
			If true, subtracts any children genotypes from the parent's frequency. Used to make the muller diagrams more intuitive.
			This will make the muller plot show each genotype that reaches 100% as actually taking up the full vertical space in the graphic,
			Rather than as 50%/50% alongside any child genotypes.
	"""

	def __init__(self, cutoff_detection: float, adjust_populations: bool = True):
		self.cutoff_detection = cutoff_detection
		self.adjust_populations = adjust_populations

		# This is the amount to preserve when adjusting the populations frequencies to account for child genotypes.
		# This will allow the parent genotypes to essentially "outline" the child genotypes.
		self.visible_slice = 0.01
		self.ancestral_genotype_label = 'genotype-0'  # The name to assign to the ancestral genotype.

	@staticmethod
	def _compile_parent_linkage(edges: pandas.Series) -> Dict[str, List[str]]:
		""" Maps a genotype to a list of all genotypes that inherit from it"""
		children = dict()
		for identity, parent in edges.items():
			if parent not in children:
				children[parent] = list()
			children[parent].append(identity)
		return children

	@staticmethod
	def _convert_genotype_table_to_population_table(genotype_table: pandas.DataFrame) -> pandas.DataFrame:
		""" Pivots a genotype table into a long-form table.
			Parameters
			----------
			genotype_table: pandas.DataFrame
				 The genotype table to pivot.

			Returns
			-------
			pandas.DatFrame
				- columns
					`Generation`: int
					`Identity`: str
					`Population`: float
		"""
		table = list()
		for genotype_label, genotype_frequencies in genotype_table.iterrows():
			# Remove timepoints where the value was 0 or less than 0 due to above line.
			for timepoint, frequency in genotype_frequencies.items():
				row = {
					'Identity':   genotype_label,
					'Generation': int(timepoint),
					'Population': frequency * 100
				}
				table.append(row)

		temp_df = pandas.DataFrame(table)
		return temp_df

	def _subtract_children_from_parent(self, modified_genotypes: pandas.DataFrame, children: Dict[str, List[str]]) -> pandas.DataFrame:
		""" Reduces the observed frequency of parent genotypes to allow child genotypes to be visible at the correct vertical abundance."""

		children_table = list()
		for genotype_label, genotype in modified_genotypes.iterrows():
			if genotype_label in children:
				genotype_frequencies = genotype[genotype > self.cutoff_detection]
				genotype_children: pandas.Series = modified_genotypes.loc[children[genotype_label]].max()
				genotype_frequencies: pandas.Series = genotype_frequencies - genotype_children
				genotype_frequencies = genotype_frequencies.mask(lambda s: s < self.cutoff_detection, self.visible_slice)
				genotype_frequencies = genotype_frequencies.fillna(0)  # Otherwise plotting the muller diagram will fail.
			else:
				genotype_frequencies = genotype
			genotype_frequencies.name = genotype_label
			children_table.append(genotype_frequencies)
		return pandas.DataFrame(children_table)

	def add_ancestral_genotype(self, population_table: pandas.DataFrame) -> pandas.DataFrame:
		""" Adds the ancestral genotype to the graphic. This is based on the observed frequencies at each timepoint and whether they sum to 100%."""
		generation_groups = population_table.groupby(by = 'Generation')
		modified_population = population_table
		for generation, group in generation_groups:
			# Check if the observed genotypes collectively sum to 100%.
			population = group['Population'].sum()
			if population > 100:
				p = 0
			else:
				p = 100 - population
			current_timepoint = {
				'Generation': generation,
				"Identity":   self.ancestral_genotype_label,
				"Population": p
			}
			# Since `modified_population` is a dataframe, need to manually tell it to ignore the index.
			modified_population = modified_population.append(current_timepoint, ignore_index = True)
		return modified_population

	def generate_ggmuller_population_table(self, edges: pandas.Series, mean_genotypes: pandas.DataFrame) -> pandas.DataFrame:
		"""
			Converts the genotype frequencies to a population table suitable for ggmuller.
		Parameters
		----------
		mean_genotypes: pandas.DataFrame
			Contains the mean frequency for each genotype at each timepoint. Assumes that the table is indeed by
			genotype label and only the timepoint columns are included.
		edges: pandas.Series
			A Mapping from child genotype to parent genotype.

		Returns
		-------
		pandas.DataFrame
		"""

		# "Generation", "Identity" and "Population"
		# Adjust populations to account for inheritance.
		# If a child genotype fixed, the parent genotype should be replaced.

		# Use a copy of the dataframe to avoid making changes to the original.
		modified_genotypes: pandas.DataFrame = mean_genotypes.copy(True)
		# In case the genotype table includes genotypes that the edges table does not have.
		modified_genotypes = modified_genotypes[modified_genotypes.index.isin(edges.index)]

		# Generate a list of all genotypes that arise in the background of each genotype.
		# Should ba a dict mapping parent -> list[children]
		children = self._compile_parent_linkage(edges)
		if self.adjust_populations:
			child_df = self._subtract_children_from_parent(modified_genotypes, children)
		else:
			child_df = modified_genotypes
		temp_df = self._convert_genotype_table_to_population_table(child_df)
		population_table = self.add_ancestral_genotype(temp_df)

		return population_table


def generate_trajectory_table(trajectories: pandas.DataFrame, parent_genotypes: Union[pandas.Series, Dict[str, str]],
		info: pandas.DataFrame) -> pandas.DataFrame:
	# Sorts the trajectory table and adds additional columns from the original table.
	trajectories = trajectories[sorted(trajectories.columns, key = lambda s: int(s))]
	if info is not None:
		trajectory_table: pandas.DataFrame = trajectories.copy()
		trajectory_table['genotype'] = [parent_genotypes.get(k) for k in trajectory_table.index]
		trajectory_table = trajectory_table.join(info).sort_values(by = ['genotype'])
		return trajectory_table



"""
	Python implementation of the Muller_plot function available from ggmuller.
"""

import math
import random
from itertools import filterfalse
from pathlib import Path
from typing import *

import pandas
from loguru import logger
from matplotlib import pyplot as plt
# plt.switch_backend('agg')
from matplotlib.figure import Axes  # For autocomplete

# Make sure the svg files save the labels as actualt text.
plt.rcParams['svg.fonttype'] = 'none'
try:
	from muller.widgets import calculate_luminance
	from muller.graphics import Palette
except (ModuleNotFoundError, ImportError):
	from ..widgets import calculate_luminance
	from .palettes import Palette

NumericArrayType = List[Union[int,float]]
plt.style.use('seaborn-white')

pandas.set_option('display.max_rows', 500)
pandas.set_option('display.max_columns', 500)
pandas.set_option('display.width', 250)


###################################################################################################################################################
################################################################# Coordinates #####################################################################
###################################################################################################################################################


def unique_everseen(iterable: Iterable[Any], key: Optional[Callable[[Any], str]] = None) -> Generator[Any, None, None]:
	"""	List unique elements, preserving order. Remember all elements ever seen."
		unique_everseen('AAAABBBCCDAABBB') --> A B C D
		unique_everseen('ABBCcAD', str.lower) --> A B C D
	"""
	seen = set()
	seen_add = seen.add
	if key is None:
		for element in filterfalse(seen.__contains__, iterable):
			seen_add(element)
			yield element
	else:
		for element in iterable:
			k = key(element)
			if k not in seen:
				seen_add(k)
				yield element


def distance(left: Tuple[float, float], right: Tuple[float, float]) -> float:
	xl, yl = left
	xr, yr = right

	num = (xl - xr) ** 2 + (yl - yr) ** 2
	return math.sqrt(num)


def find_closest_point(point, points: List[Tuple[float, float]]) -> Tuple[float, float]:
	return min(points, key = lambda s: distance(s, point))


def relocate_point(point: Tuple[float, float], locations: List[Tuple[float, float]]) -> Tuple[float, float]:
	x_loc, y_loc = point
	for _ in range(10):
		closest_neighbor = find_closest_point((x_loc, y_loc), locations)
		y_is_close = math.isclose(y_loc, closest_neighbor[1], abs_tol = .1)
		y_is_almost_close = math.isclose(y_loc, closest_neighbor[1], abs_tol = .3)
		x_is_close = math.isclose(x_loc, closest_neighbor[0], abs_tol = 2)
		x_large = x_loc >= closest_neighbor[0]
		if distance(closest_neighbor, (x_loc, y_loc)) > 1: break
		if y_is_close:
			if y_loc > closest_neighbor[1]:
				y_loc += random.uniform(.1, .2)
			else:
				y_loc -= random.uniform(.05, .10)
		elif y_is_almost_close and x_is_close and x_large:
			x_loc += .5
			break
		else:
			break
	return x_loc, y_loc


###################################################################################################################################################
################################################################# Graphics #####################################################################
###################################################################################################################################################
class MullerPlot:
	"""
		Workflow for generating a muller diagram.
		Parameters
		----------
		outlines: bool
			Whether to add an outline each genotype.
		render: bool
			Whther to generate an svg or pdf render of the diagram.
		style: Literal['default']
			Specifies one of the pre-defined styles to use when generating the plots.
		smooth_plot:bool; default True
			Whether to smooth out the timeseries data when plotting.
	"""

	def __init__(self, outlines: bool, render: bool, style: Optional[str] = 'default', scale: int = 1):
		self.outlines = outlines
		self.render = render
		self.dpi = 250

		self.root_genotype_name = 'genotype-0'
		self.outline_color = '#FFFFFF' if self.outlines else None  # The color of the genotype outline
		self.root_genotype_color = '#FFFFFF'  # The color of the root genotype.
		self.style = style

		# Properties for the size of the plot labels.

		# Properties for the genotype annotations that will be added to the plot
		self.genotype_annotation_outline_color = "#FFFFFF"
		self.genotype_annotation_outline_linewidth = 0.5
		self.genotype_annotation_alpha = 1

		self.genotype_annotation_font_size = 9  # Size of the text in the genotype annotation box.
		self.genotype_annotation_font_color_light = "#FFFFFF"  # Font color to use when background is dark
		self.genotype_annotation_font_color_dark = "#333333"  # Font color to use when the background is light.

		self.scale = scale
		self.reference_x = 20
		self.reference_y = 20
		self.scale_max = 3  # The maximum scale that the plots can be scalled to in order to avoid image limitations.

		# Set up the fontsizes for each labeltype
		self.label_size_axis, self.label_size_title, self.label_size_ticks = self.set_scale(scale)

		# Set the interpolation method.
		# This makes the plots look a little less jarring.
		# Available options: ‘linear’, ‘nearest’, ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’
		self.interpolation_type = 'slinear'
		# Set the number of points to generate with the interpolation method
		self.interpolation_size = 100

	def _apply_style(self, ax: plt.Axes, title: Optional[str], maximum_x: float):
		"""
			Addes labels and such to the graph.
		Parameters
		----------
		ax: THe plt.Axes object the graph was plotted on.
		maximum_x: The maximum x-value observed in the dataset.
		"""
		if self.style == 'nature':
			self._set_style_nature()
		else:
			self._set_style_default()

		ax.set_facecolor(self.root_genotype_color)

		# Add labels to the x and y axes.
		if title:
			ax.set_title(title, fontsize = self.label_size_title)
		ax.set_xlabel(self.xaxis_label, fontsize = self.label_size_axis)
		ax.set_ylabel(self.yaxis_label, fontsize = self.label_size_axis)
		# Increase the font size of the frequency and timepoint labels
		ax.tick_params(axis = 'both', labelsize = self.label_size_ticks)
		# Basic stacked area chart.

		# Remove extra ticks
		if self.style == 'nature':
			ax.set_yticks([0, 0.5, 1])
		else:
			ax.spines['right'].set_visible(False)
			ax.spines['top'].set_visible(False)

		# Make sure that the width of the diagram only shows timepoints we have data for.
		ax.set_xlim(0, maximum_x)
		ax.set_ylim(0, 1)

		return ax

	def _initialize_plot(self, ax: plt.Axes) -> plt.Axes:
		"""
			A typical plot will have about 10 to 20 genotypes and 10-20 timepoints. The current defaults work well for
			datasets of that size, but need to scale the plot size for anything larger.
		"""
		if ax is None:
			figsize_x = self.scale * 12
			figsize_y = self.scale * 10

			fig, ax = plt.subplots(figsize = (figsize_x, figsize_y))
		return ax

	def _get_annotation_label_background_properties(self, genotype_color: str) -> Dict[str, str]:
		""" Returns a dictionary with the properties describing the genotype annotation background."""
		background_properties = {
			'facecolor': genotype_color,
			'alpha':     self.genotype_annotation_alpha,
			'edgecolor': self.genotype_annotation_outline_color,
			'linewidth': self.genotype_annotation_outline_linewidth
		}
		return background_properties

	def _get_annotation_label_font_properties(self, genotype_color: str) -> Dict[str, Any]:
		"""
			Generates font properties for each annotations. The main difference is font color,
			which is determined by the background color.
		Parameters
		----------
		genotype_color: str
			The color of the genotype this text is associated with.

		Returns
		-------
		fontproperties: Dict[str, Any]
		"""

		luminance = calculate_luminance(genotype_color)
		label_properties = {
			'size':  self.genotype_annotation_font_size,
			'color': self.genotype_annotation_font_color_dark if luminance > 0.5 else self.genotype_annotation_font_color_light
		}
		return label_properties

	@staticmethod
	def _merge_groups(left: pandas.DataFrame, right: pandas.DataFrame) -> pandas.DataFrame:
		"""
			Merges two genotype series which are ajacent to each other.
		"""
		left = left.set_index('Generation')
		right = right.set_index('Generation')
		frequencies = left['Frequency'].add(right['Frequency'], fill_value = 0)
		populations = left['Population'].add(right['Population'], fill_value = 0)

		df = pandas.DataFrame([frequencies, populations]).transpose()
		# Add the other columns to the dataframe
		reference = left.iloc[0]
		df['Group_id'] = reference['Group_id']
		df['Identity'] = reference['Identity']
		df['Unique_id'] = [f"{i}_{j}" for i, j in zip(df['Identity'].values, df.index)]

		df = df.reset_index()
		return df

	def _set_style_default(self):
		self.xaxis_label = 'Generation'
		self.yaxis_label = 'Frequency'
		self.root_genotype_color = "white"  # More readable that '#FFFFFF'

	def _set_style_nature(self):
		self.xaxis_label = 'Generation'
		self.yaxis_label = 'Individuals'
		self.root_genotype_color = 'white'

	def _set_scaling_factor(self, width_x: int, width_y: int) -> int:
		""" Since very large datasets require the plots to scale up, try to infer a decent scaling factor
			So that the resulting plots are actually legible.
		"""
		scale_x = round(width_x / self.reference_x)
		scale_y = round(width_y / self.reference_y)
		# Make sure they aren't 0
		if scale_x == 0: scale_x = 1
		if scale_y == 0: scale_y = 1

		if scale_x > self.scale_max:
			logger.debug(f"Timeseries Plot: Reducing x scale from {scale_x} to {self.scale_max}.")
			scale_x = self.scale_max
		if scale_y > self.scale_max:
			logger.debug(f"Timeseries Plot: Reducing y scale from {scale_y} to {self.scale_max}.")
			scale_y = self.scale_max
		self.scale = max(scale_x, scale_y)
		return self.scale

	@staticmethod
	def set_scale(scale: int = 1) -> Tuple[int, int, int]:
		# Made staticmethod so that pycharm doesn't complain about object properties being defined outside of __init__()
		label_size_axis = 24 * scale
		label_size_title = 42 * scale
		label_size_ticks = 18 * scale
		return label_size_axis, label_size_title, label_size_ticks

	@staticmethod
	def get_default_palette(labels: Iterable[str]) -> Dict[str, str]:
		from . import palettes

		return palettes.generate_palette(labels)

	def save_figure(self, filename: Path):
		""" Saves the diagram in every format available in self.filetypes"""
		plt.savefig(filename, dpi = self.dpi if filename.suffix != '.svg' else None)  # Not sure if setting DPI for svgs raises an error.

	@staticmethod
	def generate_muller_series(muller_df: pandas.DataFrame, color_palette: Dict[str, str]) -> Tuple[
		List[float], List[List[float]], List[str], List[str]]:
		"""
			Generates the required inputs for matplotlib to generate a lineage.
		Parameters
		----------
		muller_df: pandas.DataFrame
		color_palette: Dict[str,str]
			Maps genotypes to a specific color.

		Returns
		-------
		x, y, colors, labels
		"""
		genotype_order = list(unique_everseen(muller_df['Group_id'].tolist()))

		x_values = list(unique_everseen(muller_df['Generation'].tolist()))
		# Extract the labels for each genotype, preserving the order they will be plotted in. If the
		# genotype was split up, only keep one of the series labels.

		# labels = [(label if not label.endswith('a') else None) for label in genotype_order]
		labels = muller_df['Identity'].unique()
		colors = [color_palette[label[:-1] if (label.endswith('a') and label not in labels) else label] for label in genotype_order]

		groups = muller_df.groupby(by = 'Group_id')

		# Keep track of which genotype series have already been processed.
		# Skip the second genotype series if both halves have already been merged.
		ys = list()
		for index, label in enumerate(genotype_order):
			# The genotype contains child genotypes, so it has to be kept split.
			genotype_series = groups.get_group(label)['Frequency'].tolist()
			ys.append(genotype_series)

		return x_values, ys, colors, labels

	def merge_muller_table(self, muller_df: pandas.DataFrame) -> pandas.DataFrame:
		""" The muller dataframe splits genotypes into halves so they can be plotted as a normal stacked area chart.
			We want to merge these halves if they are plotted adjacent to each other.
		"""
		# Get the order that each series appears.
		genotype_order = list(unique_everseen(muller_df['Group_id'].tolist()))
		# Each genotype should have a corresponding second-half table.
		groups = muller_df.groupby(by = 'Group_id')
		seen = set()
		merged_table = list()
		# Iterate over the population table and merge adjacent genotype subseries.
		for index, label in enumerate(genotype_order):
			if label in seen: continue
			seen.add(label)
			# Check if both of the series for this genotype are neighbors. If so, merge them.
			try:
				next_label = genotype_order[index + 1]
			except IndexError:
				next_label = "   "  # Three spaces, so the following subscription works
			# Check whether the genotype label is the same, and that the second label has an additional `a` character at the end.
			if label == next_label[:-1] and next_label[-1].isalpha():  # make sure it ends with a letter.
				# The two halves of this specific genotype are next to each other. combine them and skip the next iteration.
				seen.add(next_label)
				genotype_series = self._merge_groups(groups.get_group(label), groups.get_group(next_label))
			else:
				# The genotype contains child genotypes, so it has to be kept split.
				genotype_series = groups.get_group(label)
			merged_table.append(genotype_series)
		return pandas.concat(merged_table, sort = True)

	def add_genotype_annotations_to_plot(self, ax: Axes, points: Dict[str, Tuple[float, float]], annotations: Dict[str, List[str]],
			color_palette: Dict[str, str]) -> Axes:
		locations = list()
		for genotype_label, point in points.items():
			if genotype_label == self.root_genotype_name:
				# There's no point in adding an annotation for the root genotype.
				continue

			genotype_color: str = color_palette[genotype_label]
			genotype_annotations: List[str] = annotations.get(genotype_label, [])
			if not genotype_annotations:
				# No annotations for this genome. Don't draw anything.
				continue

			background_properties = self._get_annotation_label_background_properties(genotype_color)
			label_properties = self._get_annotation_label_font_properties(genotype_color)

			if locations:
				x_loc, y_loc = relocate_point(point, locations)
			else:
				x_loc, y_loc = point

			locations.append((x_loc, y_loc))
			ax.text(
				x_loc, y_loc,
				"\n".join(genotype_annotations),
				bbox = background_properties,
				fontdict = label_properties
			)
		return ax

	def get_coordinates(self, muller_df: pandas.DataFrame) -> Dict[str, Tuple[int, float]]:
		"""
			Attempts to find the mean point of the stacked area plot for each genotype.
			This is used to determine where to draw any annotations available for the given genotype.
		Parameters
		----------
		muller_df:pandas.DataFrame

		Returns
		-------
		points: Dict[str, Tuple[int, float]]
		A dictionary mapping each genotype to an estimate of the midpoint of the genotype's area.
		"""

		# We need to keep track of when each series is plotted. There will be one or two series for each genotype, depending on if they have
		# been merged.
		genotype_order = list()
		for index, row in muller_df.iterrows():
			genotype_label = row['Group_id']
			if genotype_label not in genotype_order:
				genotype_order.append(genotype_label)

		# We don't care about when the series had a value of zero. Exclude these timepoints.
		nonzero = muller_df[muller_df['Population'] != 0]

		# Iterate over the nonzero timepoints for each series and try to find the middle of the series.
		points = dict()
		groups = nonzero.groupby(by = 'Group_id')

		for index, name, in enumerate(genotype_order):
			# muller_df splits each genotype so that it can draw them in the correct order as a stacked area chart.
			# The second series label has an additional 'a' character at the end to distinguish it from the first series for each genotype.
			genotype_label = name[:-1] if (name.endswith('a') and name not in muller_df['Identity'].unique()) else name

			# Check if this genotype has already been assigned a location.
			if genotype_label in points: continue
			try:
				group = groups.get_group(name)
			except KeyError:
				# Just in case something weird happens.
				continue

			# Find the mean timepoint for this series.
			centroid = self.calculate_centroid(group.set_index('Generation')['Frequency'])

			# This is the centroid for the genotype series as-is.
			# This does not take into account the fact that the previously plotted genotypes will change the plotted
			# y-value of the series. So, find how much the y-value is changed due to plotting and add it to the centroid's y-value.
			previous_genotype_labels = genotype_order[:index]
			previous_genotype_table = muller_df[(muller_df['Group_id'].isin(previous_genotype_labels)) & (muller_df['Generation'] == centroid[0])]
			previous_frequencies = sum(previous_genotype_table['Frequency'].values)
			points[genotype_label] = (centroid[0], previous_frequencies + centroid[1])
		return points

	@staticmethod
	def calculate_centroid(timeseries: pandas.Series) -> Tuple[int, float]:
		""" Attempts to find the centroid for a timeseries. Used to figure out where to plot the genotype annotation"""
		mean_x = (max(timeseries.index) - min(timeseries.index)) / 2

		if mean_x not in timeseries.index:
			# Find the closest timepoint to this mean.
			mean_x = sorted(timeseries.index, key = lambda s: s - mean_x)[0]

		mean_y = timeseries.loc[mean_x]

		return mean_x, mean_y

	def interpolate_values(self, x:NumericArrayType, y: NumericArrayType)->Tuple[NumericArrayType, NumericArrayType]:
		""" Interpolates the values of the input arrays. """
		from scipy import interpolate

		interpolation = interpolate.interp1d(x, y, kind = self.interpolation_type)

		# Generate the interolated values
		x_value_minimum = min(x)
		x_value_maximum = max(x)
		step = (x_value_maximum - x_value_minimum) / self.interpolation_size  # The increment number for the new x-axis.
		x_interpolated = [x_value_minimum + (step * i) for i in range(self.interpolation_size)]
		y_interpolated = [interpolation(i) for i in x_interpolated]

		return x_interpolated, y_interpolated




	def plot(self, muller_df: pandas.DataFrame, filename:Path = None, color_palette: Dict[str, str] = None,
			annotations: Optional[Dict[str, List[str]]] = None,
			title: Optional[str] = None, ax: Optional[plt.Axes] = None) -> plt.Axes:
		"""
			Generates a muller diagram equivilent to r's ggmuller. The primary advantage is easier annotations.
		Parameters
		----------
		muller_df: pandas.DataFrame
		color_palette: Dict[str,str]
		filename: Path
		annotations: Dict[str, List[str]]
			A map of genotype labels to add to the plot.
		title: Optional[str]
			Applies the given title to the plot
		ax: plt.Axes
			A preexisting Axes object to draw the plot on.
		Returns
		-------
		ax: Axes
			The axes object containing the plot.
		"""

		if color_palette is None:
			color_palette = self.get_default_palette(muller_df['Identity'].unique())

		points = self.get_coordinates(muller_df)

		merged_table = self.merge_muller_table(muller_df)

		# Need to scale the graph so very large datasets are still easily distinguishable.
		# Calculate what the scale should be given the dataset shape. Any dataset smaller than the reference is automatically converted to 1.
		self.set_scale()
		# Disable the white outlines if the dataset is very large.
		if self.scale > 2:
			self.outline_color = None  # Disable the white outlines.

		ax = self._initialize_plot(ax)
		x, y, colors, labels = self.generate_muller_series(merged_table, color_palette)
		y_interpolated = list()
		for y_series in y:
			x_interpolated, yi = self.interpolate_values(x, y_series)
			y_interpolated.append(yi)

		#x_interpolated, y_interpolated = self.interpolate_values(x, y)

		ax.stackplot(
			x_interpolated, y_interpolated,
			colors = colors,
			labels = labels,
			edgecolor = self.outline_color,
			linewidth = 2,
			interpolate = True,
			joinstyle = 'round'
		)

		if annotations:
			self.add_genotype_annotations_to_plot(ax, points, annotations, color_palette)

		ax = self._apply_style(ax, title, max(x))
		if filename:
			#logger.info(f"Saving the muller plot as {filename.absolute()}")
			self.save_figure(filename)

		return ax



from typing import List, Optional

import pandas
from loguru import logger


class TrajectoryFilter:
	""" Filters trajectories (not genotypes) based on a set of criteria aimed at removing erroneous measurements.
		Parameters
		----------
		detection_cutoff,fixed_cutoff: float
			The detection/fixed cutoffs for this population.
		filter_consistency: float
			The amount by which a mutational trajectory must change in order to not be filtered out.
		filter_single: bool
			Whether to filter out mutational trajectories which only exist at a single timepoint.
		filter_startfixed: bool
			Whether to filter out mutational trajectories which begin the experiment fixed.
	"""

	def __init__(self, detection_cutoff: float, fixed_cutoff: float, filter_consistency: float = 0.1, filter_single: bool = True,
			filter_startfixed: bool = True):
		self.dlimit: float = detection_cutoff
		self.flimit: float = fixed_cutoff
		self.filter_consistency: float = filter_consistency
		self.use_filter_single: float = filter_single
		self.use_filter_startfixed: bool = filter_startfixed

	def run(self, trajectory_table: pandas.DataFrame) -> pandas.DataFrame:
		"""
			Filters out individual trajectories that fail certain filters.
		Parameters
		----------
		trajectory_table:pandas.DataFrame
		"""
		# Remove trajectories that only exist at one timepoint and exceed the fixed cutoff limit.

		filtered_trajectories = trajectory_table.apply(self.apply, axis = 1)
		logger.debug(filtered_trajectories.to_string())
		filtered_trajectories = filtered_trajectories[filtered_trajectories != "passed"]
		if len(filtered_trajectories) > 0:
			logger.warning(f"These trajectories did not pass the trajectory filters:")
			for label, reason in filtered_trajectories.items():
				logger.warning(f"\t{label}: {reason}")
		return trajectory_table[~trajectory_table.index.isin(filtered_trajectories.index)]

	def apply(self, trajectory: pandas.Series) -> str:
		"""Applies each filter to the trajectory. Returns the first filter that was not passed."""
		test_result = None
		if self.use_filter_single:
			test_result = self.trajectory_only_detected_once(trajectory)
		if test_result:
			return "onlyDetectedOnce"

		if self.use_filter_startfixed:
			test_result = self.trajectory_started_fixed(trajectory)
		if test_result:
			return "startedFixed"

		if self.filter_consistency > 0:
			test_result = self.trajectory_is_constant(trajectory)
		if test_result:
			return "isConstant"
		return 'passed'
	def trajectory_started_fixed(self, trajectory: pandas.Series)->bool:
		return trajectory.iloc[0] > self.flimit

	def trajectory_only_detected_once(self, trajectory: pandas.Series) -> bool:
		within_range: List[bool] = [(self.dlimit < s) for s in trajectory.values]  # faster than using .apply()
		return sum(within_range) < 2

	def trajectory_is_constant(self, trajectory: pandas.Series) -> bool:
		"""
			Determines whether a specific trajectory remains at a reletively constant frequency throughout the experiment.
			Trajectories must change in frequency by at least 10% over the course of the experiment.
		"""

		maximum_difference = trajectory.max() - trajectory.min()

		return maximum_difference <= self.filter_consistency


class GenotypeFilter:
	""" Filters genotypes (not trajectories) based on a set of criteria aimed at removing genotypes which do not make sense
		in the context of an evolutionary experiment. Mainly, genotypes which fix should wipe out all other genotypes not contained
		within the fixed background.
	"""

	def __init__(self, detection_cutoff: float, fixed_cutoff: float, frequencies: List[float], strict: bool = False):
		self.detection_cutoff = detection_cutoff
		self.fixed_cutoff = fixed_cutoff
		self.frequencies = frequencies
		self.strict = strict
		self.filtered_trajectories: List[str] = []
		self.fuzzy_fixed_cutoff = fixed_cutoff  # Should be updated in the `get_fuzzy_backgrounds` method.

	def run(self, genotypes: pandas.DataFrame, genotype_members: pandas.Series) -> List[str]:
		"""
			Finds all trajectories which should be filtered out of the dataset based on certain criteria.
			 These criteria concern both the trajectories and their parent genotypes.
		Parameters
		----------
		genotypes: pandas.DataFrame
			pre-computed genotypes table.
		genotype_members: pandas.Series
			Maps genotypes to a '|' delimited string of member trajectories.

		Returns
		-------
		List[str]
			A list of all trajectories which should be filtered out of the dataset.
		"""

		invalid_genotype = self.filter_genotype(genotypes)

		if invalid_genotype:
			# Get a list of the trajectories that form this genotype.
			invalid_members = genotype_members.loc[invalid_genotype].split('|')
			logger.info(f"A genotype consisting of trajectories ({invalid_members}) failed the genotype filters: " + str(invalid_members)[1:-1])
			return invalid_members
		else:
			return []

	def filter_genotype(self, genotypes: pandas.DataFrame) -> Optional[str]:
		""" Returns the label of the first genotype which fails the filtering criteria."""
		current_backgrounds = self.get_fuzzy_backgrounds(genotypes)
		logger.debug("Backgrounds for filtering:")
		for background, row in current_backgrounds.iterrows():
			logger.debug(f"\t{background}\t{max(row)}")

		# Search for genotypes that do not make sense in the context of an evolved population.
		current_invalid_genotype = self.find_first_invalid_genotype(genotypes, current_backgrounds)

		return current_invalid_genotype

	def get_fuzzy_backgrounds(self, genotypes: pandas.DataFrame) -> pandas.DataFrame:
		""" Extracts the backgrounds using a list of frequency breakpoints and sets the `fuzzy_fixed_cutoff` attribute."""
		for cutoff in self.frequencies:
			backgrounds = genotypes[genotypes.max(axis = 1) > cutoff]
			# Assume that backgrounds with a single timepoint are filtered during the filtering step.
			#backgrounds = backgrounds[backgrounds.sum(axis = 1) > 2*cutof]  # Check if background appears at multiple timepoints.
			if not backgrounds.empty:
				fuzzy_fixed_cutoff = cutoff
				break
		else:
			raise ValueError("The filters cannot be applied since no backgrounds can be detected.")
		self.fuzzy_fixed_cutoff = fuzzy_fixed_cutoff
		return backgrounds

	def find_first_invalid_genotype(self, genotypes: pandas.DataFrame, backgrounds: pandas.DataFrame) -> Optional[str]:
		"""	Invalid genotypes are those that don't make sense in the context of evolved populations. For example, when a genotype fixes it wipes out all
			unrelated diversity and essentially 'resets' the mutation pool. Genotypes which are detected prior to a fixed genotype should, in theory,
			fall to an undetected frequency. Any genotypes that do not follow this rule (are detected both before and after a genotype fixes) should
			be considered invalid and removed from the population. The genotypes then should be re-calculated with the offending trajectories removed.
		Parameters
		----------
		genotypes: pands.DataFrame
		backgrounds: pandas.DataFrame
		"""

		# Filter out backgrounds that are only detected at one timepoint.
		# backgrounds = _get_backgrounds_present_at_multiple_timepoints(backgrounds, detection_cutoff)
		# We want to iterate over the non-background genotypes to check if they appear both before and after any genotypes that fix.
		not_backgrounds = genotypes[~genotypes.index.isin(backgrounds.index)]
		# Iterate over the detected backgrounds.
		for _, background in backgrounds.iterrows():
			# Find the timepoint where the background first fixes.
			first_detected_point: int = self.get_first_timepoint_above_cutoff(background, self.detection_cutoff)
			first_fixed_point: int = self.get_first_timepoint_above_cutoff(background, self.fuzzy_fixed_cutoff)
			background_value_at_first_fixed_point = background.loc[first_fixed_point]
			# Iterate over the non-background genotypes.
			for genotype_label, genotype in not_backgrounds.iterrows():
				# Double check that it is not a background
				if genotype_label in backgrounds.index: continue
				# Check if it is invalid.
				is_invalid = self.check_if_genotype_is_invalid(genotype, first_detected_point, first_fixed_point,background_value_at_first_fixed_point)
				if is_invalid is not None:
					return genotype_label

	# noinspection PyTypeChecker
	@staticmethod
	def get_first_timepoint_above_cutoff(series: pandas.Series, cutoff: float) -> int:
		""" Extracts the first timepoint that exceeds the cutoff."""
		return series[series > cutoff].idxmin()

	# noinspection PyTypeChecker
	def check_if_genotype_is_invalid(self, genotype: pandas.Series, background_detected: int, background_fixed: int, background_value: float) -> Optional[str]:
		"""
			Checks if a genotype does not adhere to certain assumptions related to the background.
			1. The genotype should not be present both before and after a background fixes, and should be nonzero when the background fixes.
		Parameters
		----------
		genotype: pandas.Series
			The genotype being tested.
		background_detected: int
			The first timepoint the background was detected.
		background_fixed: int
			The first timepoint the background qualifies as 'fixed'
		background_value: float
			The value of the background when it is considered fixed. This is needed since we are using fuzzy cutoffs to find the most abundant genotype.

		Returns
		-------
		bool
		"""
		# get the nonzero timepoints for the genotype.
		detected_points = genotype[genotype > self.detection_cutoff]

		if len(detected_points) < 2:
			# The genotype is either undetected at all timepoints or is detected once.
			return "notEnoughTimpoints"

		first_detected = detected_points.first_valid_index()
		last_detected = detected_points.last_valid_index()

		# Check if the genotype was detected prior to the background. Skip those that were not.
		was_detected_before_and_after_background = first_detected < background_detected < last_detected
		# Check if the genotype was detected before and after the timepoint the current backgound fixed.
		was_detected_before_and_after_fixed = first_detected < background_fixed < last_detected
		# print(genotype.name, (first_detected, background_detected), (last_detected, background_detected), was_detected_before_and_after_fixed, was_detected_before_and_after_background)
		if was_detected_before_and_after_background and was_detected_before_and_after_fixed:
			# To confirm that it is an invalid genotype rather than a genotype that was wiped out by a background and then reapeared,
			# Check to see if it was undetected at the timpont the background fixed.
			value_at_fixed_point = genotype.loc[background_fixed]
			fixed_point_value = value_at_fixed_point + background_value
			if self.strict or (value_at_fixed_point > self.detection_cutoff and fixed_point_value > (1 + self.detection_cutoff)):
				# The genotype was detected at the first fixed timepoint.
				return "presentBeforeAndAfterFixedGenotype"
		return None



import re
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import pandas
from loguru import logger

try:
	from muller.clustering import metrics, genotype_reorder, hierarchy
	from muller import filters
	from muller.dataio import projectdata
except ModuleNotFoundError as exception:
	logger.warning(str(exception))
	from .. import filters
	from . import metrics, hierarchy
def is_trajectory_labeled_by_genotype(label):
	regex = "trajectory-[a-z]+-[0-9]+"
	match = re.search(regex, label)
	return match

def generate_genotype_name(index: int, members: List[str]):
	""" Generates the genotype names.
		When the trajectories are named like "trajectory-aqua-2", the genotype will be named "genotype-aqua".
	"""
	# Check if the trajectories follow the naming convention
	trajectory_label_results = [is_trajectory_labeled_by_genotype(m) for m in members]
	all_trajectories_follow_convention = all(trajectory_label_results)
	try:
		unique_genotype_labels = sorted(set(i.split('-')[1] for i in members))
	except IndexError:
		unique_genotype_labels = list()
	all_trajectories_specify_same_genotype = len(unique_genotype_labels) == 1

	if all_trajectories_follow_convention and all_trajectories_specify_same_genotype:
		genotype_name = unique_genotype_labels[0]
	else:
		genotype_name = str(index)
	genotype_label = f"genotype-{genotype_name}"
	return genotype_label

class ClusterMutations:
	"""
	Parameters
	----------
	dlimit, flimit:float
		The detection, and fixed cutoff values, respectively
	metric: {'binomial', 'pearson', 'minkowski'}
		The distance metric to determine trajectory similarity.
	starting_genotypes: List[List[str]]
		A list of genotypes to start the clustering algorithm with. The distance metrics will be modified so that the trajectories specified are
		grouped together.
	"""

	def __init__(self, metric: str, dlimit: float, flimit: float,
			starting_genotypes: Optional[List[List[str]]] = None, threads: Optional[int] = None):
		self.metric: str = metric
		self.dlimit: float = dlimit
		self.flimit: float = flimit
		self.known_genotypes: List[List[str]] = starting_genotypes if starting_genotypes else []
		self.filename_pairwise = None
		self.pairwise_distances_full = None # overwritten in self.get_pairwise_distances.

		#self.filename_pairwise = filename_pairwise # Could be used to reuse a table of pairwise distances.
		# The `breakpoints` value is a bit arbitrary, so it should be safe to hard-code it.
		# 	This will actually prevent the most common error when sorting genotypes (i.e. no breakpoints given) so it's worth
		#	hard-coding it to prevent that issue.
		self.breakpoints = [1.0, 0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.2, 0.1, 0.0]

		self.distance_calculator = metrics.DistanceCalculator(
			detection_limit = self.dlimit,
			fixed_limit = self.flimit,
			metric = self.metric,
			threads = threads
		)

		self.clusterer = hierarchy.HierarchalCluster()

		self.organizer = genotype_reorder.SortGenotypeTableWorkflow(
			dlimit = dlimit,
			flimit = flimit
		)

	@staticmethod
	def _calculate_mean_frequencies_of_trajectories(name: str, genotype_timeseries: pandas.DataFrame, genotype: List[str]) -> pandas.Series:
		"""
			Generates a mean timeseries for a genotype given the timeseries of member trajectories.
		Parameters
		----------
		name: str
			The name to assign to this genotype. Usually just 'genotype-' forllowed by an integer.
		genotype_timeseries: pandas.DataFrame
			A dataframe of member trajectory timeseries.
		genotype: List[str]
			A list of the labels of all member trajectories. Used to track the which trajectories belong to this genotype.

		Returns
		-------
		pandas.DataFrame
			The calculated mean of all member trajectories at each timepoint.
		"""
		mean_genotype_timeseries = genotype_timeseries.mean()
		mean_genotype_timeseries.index = [int(i) for i in mean_genotype_timeseries.index]
		mean_genotype_timeseries['members'] = "|".join(map(str, genotype))
		mean_genotype_timeseries.name = name

		return mean_genotype_timeseries

	@staticmethod
	def _load_pairwise_distances(filename: Path) -> Dict[Tuple[str, str], float]:
		""" Reads pre-computed pairwise distances from a previous run. Typically found in the /tables/.distance.tsv table."""
		table_distance_pairwise = pandas.read_csv(filename, sep = "\t")

		pair_array = dict()
		for index, row in table_distance_pairwise.iterrows():
			values = {(index, i): row[i] for i in table_distance_pairwise.columns}
			values_inverse = {k[::-1]: v for k, v in
				values.items()}  # Faster to include reversed keys rather than trying to get (l,r) and (r,l) keys.

			pair_array.update(values)
			pair_array.update(values_inverse)

		return pair_array




	def calculate_mean_genotype(self, all_genotypes: List[List[str]], timeseries: pandas.DataFrame) -> pandas.DataFrame:
		"""
			Calculates the mean frequency of each genotype ate every timepoint.
		Parameters
		----------
		all_genotypes: List[List[str]
			A list of all muller_genotypes for a given population
		timeseries: pandas.DataFrame
			Must have a 'Trajectory' column along with the columns of the original table that represent timepoints.
			Each row corresponds to a single mutational trajectory.

		Returns
		-------
		A dataframe where every row corresponds to a genotype.
		member trajectories are listed under the 'members' column.
		every column represents a timepoint.
		"""
		mean_genotypes = list()
		for index, genotype in enumerate(all_genotypes, start = 1):
			try:
				genotype_timeseries = timeseries.loc[genotype]
			except Exception as exception:
				logger.critical(f"Missing Trajectory Labels: {genotype} - {timeseries.index}")
				raise exception
			#genotype_name = generate_genotype_name(index, genotype_timeseries.index)
			genotype_name = f"genotype-{index}"
			mean_genotype_timeseries = self._calculate_mean_frequencies_of_trajectories(genotype_name, genotype_timeseries, genotype)
			mean_genotypes.append(mean_genotype_timeseries)

		mean_genotypes = pandas.DataFrame(mean_genotypes)
		# For consistency
		mean_genotypes.index.name = 'Genotype'

		# Place the `members` column in the leftmost column
		mean_genotypes = mean_genotypes[['members'] + [i for i in mean_genotypes if i != 'members']]

		return mean_genotypes

	def generate_genotype_table(self, timepoints: pandas.DataFrame, genotypes: List[List[str]]) -> Tuple[pandas.DataFrame, Dict[str,List[str]]]:

		mean_genotypes = self.calculate_mean_genotype(genotypes, timepoints)

		# Try to keep the genotype members separate so that `mean_genotypes` is consistent.
		# The table should only have the timeseries frequency values and be indexed by genotype label.
		genotype_members = mean_genotypes.pop('members')
		# Convert the `genotype_members` pandas.Series object to a dictionary mapping genotypes to member trajectories.
		genotype_members = {k: v.split('|') for k, v in genotype_members.items()}

		return mean_genotypes, genotype_members

	def get_pairwise_distances(self, trajectories: pandas.DataFrame):
		if self.filename_pairwise:
			pair_array = self._load_pairwise_distances(self.filename_pairwise)
		else:
			pair_array = self.distance_calculator.run(trajectories)

		self.pairwise_distances_full = metrics.DistanceCache(pair_array)  # Keep a record of the pairwise distances before filtering.
		return self.pairwise_distances_full

	def run(self, trajectories: pandas.DataFrame, distance_cutoff:Optional[float] = None) -> projectdata.DataGenotypeInference:
		"""
			Run the genotype clustering workflow.
		Parameters
		----------
		trajectories: pandas.DataFrame
			A timeseries dataframe, usually generated from `import_table.import_trajectory_table`.
				- Index -> str
					Names unique to each trajectory.
				- Columns -> int
					The timeseries points will correspond to the frequencies for each trajectory included with the input sheet.
					Each trajectory/timepoint will include the observed frequency at each timepoint.
		distance_cutoff: Optional[float]
			Used to determine the distance cutoff when clustering genomtypes.
		"""

		modified_trajectories = trajectories.copy(deep = True)  # To avoid unintended changes

		# Calculate the pairwise distances between each pair of mutational trajectories.
		pairwise_distances = self.get_pairwise_distances(modified_trajectories)

		# Calculate the genotypes
		cluster_result = self.clusterer.run(
			pairwise_distances,
			starting_genotypes = self.known_genotypes,
			similarity_cutoff = distance_cutoff
		)
		genotype_table, genotype_members = self.generate_genotype_table(modified_trajectories, cluster_result.clusters)

		sorted_genotype_table = self.organizer.run(genotype_table)
		output_data = projectdata.DataGenotypeInference(
			table_trajectories = modified_trajectories,
			table_genotypes = sorted_genotype_table,
			genotype_members = genotype_members,
			matrix_distance = pairwise_distances,
			clusterdata = cluster_result,
			table_trajectories_info = None
		)

		return output_data




from pathlib import Path
from typing import Dict, List, Optional


class Palette:
	# Holds the palette used by genotypes/trajectories.
	def __init__(self, name:str, palette: Dict[str, str], members: Optional[Dict[str, List[str]]] = None):
		self.name = name
		self.palette_genotype = palette
		if members:
			self._palette_trajectory = self.generate_trajectory_palette(palette, members)
		else:
			self._palette_trajectory = None

	def __call__(self, item:str, default = None)->str:
		if item in self.palette_genotype:
			return self.palette_genotype[item]
		elif self._palette_trajectory is not None and item in self._palette_trajectory:
			return self._palette_trajectory[item]
		else:
			return default

	def __getitem__(self, item):
		return self(item)

	def generate_trajectory_palette(self, palette: Dict[str, str], members: Dict[str, List[str]]) -> Dict[str, str]:
		""" Maps the genotype palette to each of the member trajectories."""
		p = dict()
		for genotype_name, members in members.items():
			for member in members:
				p[member] = palette[genotype_name]
		self._palette_trajectory = p
		return p

	def get_genotype_palette(self) -> Dict[str, str]:
		return self.palette_genotype

	def get_trajectory_palette(self) -> Dict[str, str]:
		if self._palette_trajectory is None:
			message = f"The trajectory palette was not generated."
			raise ValueError(message)
		return self._palette_trajectory

	def get(self, item:str, default = None)->str:
		return self.palette_genotype.get(item, self._palette_trajectory.get(item, default))

	def save(self, filename:Path):
		data = {
			'name': self.name,
			'genotype': self.palette_genotype,
			'trajectory': self._palette_trajectory
		}
		import json
		filename.write_text(json.dumps(data, indent = 4, sort_keys = True))

if __name__ == "__main__":
	pass
